#include <stdio.h>
#include <math.h>
#include <gl/gl.h>
#include <gl/device.h>

#define RGB_BLACK   0x000000  
#define RGB_RED	    0x0000ff
#define RGB_MYBLUE  0xccbb88
#define RGB_GREEN   0x00ff00
#define RGB_BLUE    0xff0000
#define RGB_LIGHTBLUE    0xcc5555
#define RGB_DARKBLUE     0x665555
#define RGB_CYAN    0xffff00
#define RGB_GRAY    0xbbbbbb
#define RGB_DARKGRAY    0x999999

#define MIN(a,b) (a < b ? a : b)
#define MAX(a,b) (a > b ? a : b)

#define NUMBEROFLEGS 6

/* Definition of the robot in a given instant. */
double incxl[NUMBEROFLEGS],incyl[NUMBEROFLEGS];
double incxb,inczb;
double dtheta;

/* Motion indicators. */
double legrange;
int forward;
int rotating;
double INF = 9999999999999.0;
double xc,zc;

typedef struct {
           double x;
           double y;
           double z;
	 } vertex;

/* Definition of the actual robot. */
vertex body[8],obody[8];
vertex upper_leg[NUMBEROFLEGS][8];
vertex lower_leg[NUMBEROFLEGS][8];
vertex oupper_leg[NUMBEROFLEGS][8];
vertex olower_leg[NUMBEROFLEGS][8];

/* Definition of the "initial" body */
vertex b[8] = {
  {-1.0,0.7,0.25},
  {-1.0,0.7,-0.25},
  {1.0,0.7,-0.25},
  {1.0,0.7,0.25},
  {-1.0,0.3,0.25},
  {-1.0,0.3,-0.25},
  {1.0,0.3,-0.25},
  {1.0,0.3,0.255} };

/* Definition of the front upper left leg */
vertex ul[8] = {
  {0.55,0.55,-1.15},
  {0.65,0.55,-1.15},
  {0.7,0.6,-0.25},
  {0.5,0.6,-0.25},
  {0.55,0.45,-1.15},
  {0.65,0.45,-1.15},
  {0.7,0.4,-0.25},
  {0.5,0.4,-0.25} };


/* Definition of the front lower left leg */
vertex ll[8] = {
  {0.575,0.0,-1.225},
  {0.625,0,-1.225},
  {0.65,0.55,-1.25},
  {0.55,0.55,-1.25},
  {0.575,0,-1.175},
  {0.625,0,-1.175},
  {0.65,0.55,-1.15},
  {0.55,0.55,-1.15} };

/* Definition of the front upper right leg. */
vertex ur[8] = {
  {0.65,0.55,1.15},
  {0.55,0.55,1.15},
  {0.5,0.6,0.25},
  {0.7,0.6,0.25},
  {0.65,0.45,1.15},
  {0.55,0.45,1.15},
  {0.5,0.4,0.25},
  {0.7,0.4,0.25} };


/* Definition of the front lower right leg. */
vertex lr[8] = {
  {0.625,0.0,1.225},
  {0.575,0,1.225},
  {0.55,0.55,1.25},
  {0.65,0.55,1.25},
  {0.625,0,1.175},
  {0.575,0,1.175},  
  {0.55,0.55,1.15},
  {0.65,0.55,1.15}};

int face[6][4] = {
  {1,2,6,5},
  {5,6,7,4},
  {4,7,3,0},
  {0,3,2,1},
  {0,1,5,4},
  {2,3,7,6} };

/* Definition of the maze walls. Small, medium and 
   large. */

vertex sm[8] = {
   {-2.0,1,0.03},
   {-2.0,1,-0.03},
   {2.0,1,-0.03},
   {2.0,1,0.03},
   {-2.0,0,0.03},
   {-2.0,0,-0.03},
   {2.0,0,-0.03},
   {2.0,0,0.03} };

vertex med[8] = {
   {-3.0,1,0.03},
   {-3.0,1,-0.03},
   {3,1,-0.03},
   {3,1,0.03},
   {-3,0,0.03},
   {-3.0,0,-0.03},
   {3,0,-0.03},
   {3,0,0.03} };

vertex lar[8] = {
   {-5.0,1,0.03},
   {-5.0,1,-0.03},
   {5,1,-0.03},
   {5,1,0.03},
   {-5,0,0.03},
   {-5.0,0,-0.03},
   {5,0,-0.03},
   {5,0,0.03} };

/* Definition of the wall and maze structures. */
int NOOFWALLS;
#define MAXNOOFWALLS 100

struct Wall {
   char size;
   double theta;
   vertex t;
   short color[3];
};

struct Wall Maze[MAXNOOFWALLS];

#define deltax 0.3
#define deltay 0.3
#define deltaz 0.3

/* Here, 2 degrees. */
#define deltatheta 6 

/* Color. */
short rgbvec[3];

/* Define the viewpoint. */
vertex ILookfrom = {-3.0,4.4,0.0};
vertex ILookat = {0,0,0};
vertex Lookfrom,Lookat;

/* Define the camera viewpoint. */
vertex ICamLookfrom = {1.07,0.5,0.0};
vertex ICamLookat = {2.07,0.5,0.0};
vertex CamLookfrom,CamLookat;
vertex TCamLookfrom,TCamLookat;

double twist = 0;

/* Define the eye(center of eye ball). */
Object eye = 1;
vertex Ieye_center = {1.0,0.5,0};
vertex eye_center,Teye_center;

/* Window ids. */
long W[2];

/* Definition of the robot in a given instant. */
double incxl[NUMBEROFLEGS],incyl[NUMBEROFLEGS];
double incxb,inczb;
double dtheta;

/* Motion indicators. */
int rotating;
int moving;
int legmoving;
int movebody;
int begin_rotation;

/* Key. */
long dev;

